/*
 *
 * AMX Mod X Module
 * Basic Socket Functions
 * 
 * Codebase from Ivan, -g-s-ivan@web.de (AMX 0.9.3)
 * Modification by Olaf Reusch, kenterfie@hlsw.de (AMXX 0.16, AMX 0.96)
 * Modification by David Anderson, dvander@tcwonline.org (AMXx 0.20)
 * Modification by hackziner,  hackziner@hotmail.com
 * Modification by Destro, ema_destro_cs@hotmail.com (Only-Arg)
 *
 * Bugs/Fixes
 *
 * v0.1
 * - code structure renewed  (HLSW Dev Team)
 * v0.2
 * - added socket_send2 to send data containing null bytes (FALUCO)(AMXX v1.65) (HLSW Dev Team)
 * v0.3
 * - accept/listen/block (hackziner)
 * v0.4
 * - recvfrom/sendto (Destro)
 */

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string>

#ifdef _WIN32
#include <winsock.h>
#define socklen_t int
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#define closesocket(s) close(s)
#endif

#include "amxxmodule.h"

#define SOCKET_TCP 1
#define SOCKET_UDP 2

// native socket_open(_hostname[], _port, _protocol = SOCKET_TCP, &_error);
static cell AMX_NATIVE_CALL socket_open(AMX *amx, cell *params)  /* 2 param */
{ 
    unsigned int port = params[2];
    int len;
    char* hostname = MF_GetAmxString(amx,params[1],0,&len);
    if(len == 0)
	{
        params[4] = 2;  // server unknown
        return -1;
    }
    params[4] = 0; // params[4] is error backchannel
    struct sockaddr_in server;
    struct hostent *host_info;
    unsigned long addr;
    int sock=-1;
    int contr;

    sock = socket(AF_INET, params[3]==SOCKET_TCP?SOCK_STREAM:SOCK_DGRAM, 0);
    if(sock < 0)
	{
        // Error, couldn't create a socket, so set an error and return.
        params[4] = 1;
        return -1;
    }
    
    memset( &server, 0, sizeof (server));

    if((addr = inet_addr(hostname)) != INADDR_NONE)
	{
        memcpy((char *)&server.sin_addr, &addr, sizeof(addr));
    }
    else
	{
        host_info = gethostbyname(hostname);
        if(host_info == NULL)
		{
            params[4] = 2;  // server unknown
            return -1;
        }
        memcpy((char *)&server.sin_addr, host_info->h_addr, host_info->h_length);
    }

    server.sin_family = AF_INET;
    server.sin_port = htons(port);
    
    contr = connect(sock, (struct sockaddr*)&server, sizeof( server));
    if(contr < 0) 
	{
            params[4] = 3;  //error while connecting
            return -1;
    }

    return sock;
}

// native socket_listen(const _hostname[], _port, _protocol = SOCKET_TCP, &_error);
static cell AMX_NATIVE_CALL socket_listen(AMX *amx, cell *params)  /* 4 param */
{ 
    unsigned int port = params[2];
    int len;
    char* hostname = MF_GetAmxString(amx,params[1],0,&len);
    if(len == 0) 
	{
        params[4] = 2;  // server unknown
        return -1;
    }
    params[4] = 0; // params[4] is error backchannel
    struct sockaddr_in server;
    struct hostent *host_info;
    unsigned long addr;
    int sock=-1;
    int contr;

    sock = socket(AF_INET, params[3]==SOCKET_TCP?SOCK_STREAM:SOCK_DGRAM, 0);
    if(sock < 0)
	{
        // Error, couldn't create a socket, so set an error and return.
        params[4] = 1;
        return -1;
    }

    memset( &server, 0, sizeof (server));

    if((addr = inet_addr(hostname)) != INADDR_NONE)
	{
        memcpy((char *)&server.sin_addr, &addr, sizeof(addr));
    }
    else 
	{
        host_info = gethostbyname(hostname);
        if(host_info == NULL)
		{
            params[4] = 2;  // server unknown
            return -1;
        }
        memcpy((char *)&server.sin_addr, host_info->h_addr, host_info->h_length);
    }
    
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_family = AF_INET;
    server.sin_port = htons(port);

    bind(sock,(struct sockaddr*)&server,sizeof(server));
    
    contr=0;
    if(params[3]==SOCKET_TCP)
		listen(sock,8);

    if(contr < 0)
	{
            params[4] = 3;  //error while connecting
            return -1;
    }

    return sock;
}

// native socket_accept(_socket);
static cell AMX_NATIVE_CALL socket_accept(AMX *amx, cell *params)  /* 2 param */
{
    int NewClient;
    struct sockaddr_in cin; 
    int sinsize;
    int socket = params[1];
    NewClient = accept(socket, (struct sockaddr*)&cin, (socklen_t*)&sinsize); //hz

    if ( NewClient <0 )
       return -1;
    return NewClient;

}


// native socket_unblock(_socket);
static cell AMX_NATIVE_CALL socket_unblock(AMX *amx, cell *params)  /* 2 param */
{
    int socket = params[1];
    #if defined(O_NONBLOCK) 
	int flags; 
    if (-1 == (flags = fcntl(socket, F_GETFL, 0))) flags = 0; 
       return fcntl(socket, F_SETFL, flags | O_NONBLOCK); 
    #else 
    u_long flags = 1; 
    return ioctlsocket(socket, FIONBIO, &flags); 
    #endif
}


// native socket_close(_socket);
static cell AMX_NATIVE_CALL socket_close(AMX *amx, cell *params)  /* 2 param */
{
    int socket = params[1];

    #ifdef _WIN32
    closesocket(socket);
    #else
    close(socket);
    #endif

    return 0;
}

// native socket_change(_socket, _timeout=100000);
// 1 sec =1000000 usec
static cell AMX_NATIVE_CALL socket_change(AMX *amx, cell *params)  /* 2 param */
{
    int socket = params[1];
    unsigned int timeout = params[2];

    fd_set rfds;
    struct timeval tv;

    FD_ZERO(&rfds);
    FD_SET(socket, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = timeout;

    if(select(socket+1, &rfds, NULL, NULL, &tv)) 
       return 1; // Ok, new data, return it
     else 
       return 0; // No new data, return it
}

// native socket_recv(_socket, _data[], _length);
static cell AMX_NATIVE_CALL socket_recv(AMX *amx, cell *params)  /* 3 param */
{
    int socket = params[1];
    int length = params[3];
    int tmp = -1;

    char *tmpchar = new char[length];
    if(tmpchar == NULL)
		return -1;

    memset(tmpchar, 0, length);

    tmp = recv(socket, tmpchar, length-1, 0);
	if (tmp == -1)
	{
		delete [] tmpchar;
		return -1;
	}

    tmpchar[tmp]='\0';
    int nlen = 0;

    int max = length-1;
    const char* src = tmpchar;
    cell* dest = MF_GetAmxAddr(amx,params[2]);
    while(max--&&nlen<tmp)
	{
		*dest++ = (cell)*src++;
		nlen++;
    }
    *dest = 0;

    delete [] tmpchar;
 
    return tmp;
}

// native socket_recvfrom(_socket, _data[], _len, &_port, _ip[], _iplen);
static cell AMX_NATIVE_CALL socket_recvfrom(AMX *amx, cell *params)   /* 6 param */
{
    int socket = params[1];
    int length = params[3];
    int tmp = -1;

    char *tmpchar = new char[length];
    if(tmpchar == NULL)
		return -1;

    memset(tmpchar, 0, length);

	struct sockaddr_in cliaddr;
	int slen = sizeof(cliaddr);
    tmp = recvfrom(socket, tmpchar, length-1, 0, (struct sockaddr *)&cliaddr, &slen);
	if (tmp == -1)
	{
		delete [] tmpchar;
		return -1;
	}


    tmpchar[tmp]='\0';
    int nlen = 0;

    int max = length-1;
    const char* src = tmpchar;
    cell* dest = MF_GetAmxAddr(amx,params[2]);
    while(max--&&nlen<tmp)
	{
		*dest++ = (cell)*src++;
		nlen++;
    }
    *dest = 0;

    delete [] tmpchar;

	char ip[25];
	memcpy(ip, inet_ntoa(cliaddr.sin_addr), sizeof(ip));
	MF_SetAmxString(amx, params[5], ip, params[6]);

	cell *r_port = MF_GetAmxAddr(amx, params[4]);
	*r_port = ntohs(cliaddr.sin_port);

    return tmp;
}

// native socket_sendto(_socket, _data[], _ip[], _port);
static cell AMX_NATIVE_CALL socket_sendto(AMX *amx, cell *params)  /* 4 param */
{
	int len, len2;
	int socket = params[1];
	char* data = MF_GetAmxString(amx,params[2],0,&len);
	unsigned int port = params[4];
	char* ip = MF_GetAmxString(amx,params[3],0,&len2);

	struct sockaddr_in cliaddr;
	cliaddr.sin_family = AF_INET;
    cliaddr.sin_port = htons(port);

	unsigned long addr;

	if((addr = inet_addr(ip)) != INADDR_NONE)
	{
        memcpy((char *)&cliaddr.sin_addr, &addr, sizeof(addr));
    }
    else
	{
		struct hostent *host_info;
        host_info = gethostbyname(ip);
        if (host_info == NULL)
		{
            return -1;
        }
        memcpy((char *)&cliaddr.sin_addr, host_info->h_addr, host_info->h_length);
    }

	return sendto(socket, data, len, 0,  (struct sockaddr *)&cliaddr, sizeof(cliaddr));
}

// native socket_send(_socket, _data[]);
static cell AMX_NATIVE_CALL socket_send(AMX *amx, cell *params)  /* 2 param */
{
	int len;
	int socket = params[1];
	char* data = MF_GetAmxString(amx,params[2],0,&len);

	return send(socket, data, len, 0);
}

static char *g_buffer = NULL;
static size_t g_buflen = 0;

// native socket_send2(_socket, _data[], _length);
static cell AMX_NATIVE_CALL socket_send2(AMX *amx, cell *params)  /* 3 param */
{
	int len = params[3];
	int socket = params[1];
	if ((size_t)len > g_buflen)
	{
		delete [] g_buffer;
		g_buffer = new char[len+1];
		g_buflen = len;
	}

	cell *pData = MF_GetAmxAddr(amx, params[2]);
	char *pBuffer = g_buffer;

	while (len--)
		*pBuffer++ = (char)*pData++;

    return send(socket, g_buffer, params[3], 0);
}


AMX_NATIVE_INFO sockets_natives[] = {
	{"socket_open", socket_open},
	{"socket_close", socket_close},
	{"socket_change", socket_change},
	{"socket_recv", socket_recv},
	{"socket_send", socket_send},
	{"socket_send2", socket_send2},
	{"socket_listen", socket_listen},
	{"socket_accept", socket_accept},
	{"socket_unblock", socket_unblock},
	{"socket_recvfrom", socket_recvfrom},
	{"socket_sendto", socket_sendto},
	{NULL, NULL}
};

void OnAmxxAttach()
{
	MF_AddNatives(sockets_natives);
    #ifdef _WIN32  
        short wVersionRequested;
        WSADATA wsaData;
        wVersionRequested = MAKEWORD (1, 1);
        if (WSAStartup (wVersionRequested, &wsaData) != 0)
		{
            MF_Log("Sockets3 Error 1 (Contactar con Destro)");
        }
    #endif
    return;
}

void OnAmxxDetach()
{
    #ifdef _WIN32
    WSACleanup();
    #endif
    delete [] g_buffer;
    return;
}